/**
* GSVideo
* v0.3.3
* Changes:
* Added GPL License
* Fixed bug in loop playback.
* Author: Andres Colubri
* Processing library based on the gstremer-java bindings by Wayne Meissner:
* http://code.google.com/p/gstreamer-java/
*/

/*
  Copyright (c) 2008 Andres Colubri

  This source is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This code is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  A copy of the GNU General Public License is available on the World
  Wide Web at <http://www.gnu.org/copyleft/gpl.html>. You can also
  obtain it by writing to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package codeanticode.gsvideo;
import processing.core.*;

import java.io.*;
import java.nio.*;
import java.lang.reflect.*;

import org.gstreamer.*;
import org.gstreamer.elements.*;

/**
 *  GSMovie class.
 */
 public class GSMovie extends PImage implements PConstants {
  // no longer needing a reference to the parent because PImage has one
  //PApplet parent;

  Method movieEventMethod;
  String filename;

  boolean play;
  boolean repeat;
  boolean available;
  boolean firstFrame = true;
  int fps = 0;
  
  int movieWidth;
  int movieHeight;
  IntBuffer	moviePixels;
  
  PlayBin gplayer;
  //MainLoop gloop;
  
  public GSMovie(PApplet parent, String filename) {
    this(parent, filename, 30);
  } 
	
  public GSMovie(PApplet parent, String filename, int ifps) {
	// this creates a fake image so that the first time this
	// attempts to draw, something happens that's not an exception
	super(1, 1, RGB);

	this.parent = parent;
	gplayer = null;
    //gloop = null;
	
	File file;
	
	if (INSTANCES_COUNT == 0) {
      String[] args = { "" };
	  Gst.init("GSMovie", args);
	  //parent.println("INIT GST");
    }
	INSTANCES_COUNT++;
	
    // first check to see if this can be read locally from a file.
    try {
      try {
        // first try a local file using the dataPath. usually this will
        // work ok, but sometimes the dataPath is inside a jar file,
        // which is less fun, so this will crap out.
        file = new File(parent.dataPath(filename));
        if (file.exists()) {
          gplayer = new PlayBin("GStreamer Video Data Extractor");
          gplayer.setInputFile(file);
        }
      } catch (Exception e) { }  // ignored

      // read from a file just hanging out in the local folder.
      // this might happen when the video library is used with some
      // other application, or the person enters a full path name
      if (gplayer == null) {
        try {
          file = new File(filename);
          if (file.exists()) {
            gplayer = new PlayBin("GStreamer Video Data Extractor");
            gplayer.setInputFile(file);			
          }
        } catch (Exception e) { }
      }

	  // Network read needs to be implemented...
	  
    } catch (SecurityException se) {
      // online, whups. catch the security exception out here rather than
      // doing it three times (or whatever) for each of the cases above.
    }
	
    if (gplayer == null) {
      parent.die("Could not find movie file " + filename, null);
    }
		
    // we've got a valid movie! let's rock.
    try {
	  this.filename = filename; // for error messages
      fps = ifps;

      // register methods
      parent.registerDispose(this);

      try {
        movieEventMethod =
          parent.getClass().getMethod("movieEvent",
                                      new Class[] { GSMovie.class });
      } catch (Exception e) {
        // no such method, or an error.. which is fine, just ignore
      }
	  
	  //gloop = new MainLoop();
	  //gloop.startInBackground();
	  
      RGBDataSink videoSink = new RGBDataSink("rgb",
        new RGBDataSink.Listener() {
          public void rgbFrame(int w, int h, IntBuffer buffer) {
		    invokeEvent(w, h, buffer);
          }
        }	
      );

      gplayer.setVideoSink(videoSink); 

    } catch (Exception e) {
      e.printStackTrace();
    }

  }

  public void dispose() {
    stop();
    //gloop = null;
  } 
	
  /**
   * Get the full length of this movie (in seconds).
   */	
  public float duration() {	
    float sec = gplayer.queryDuration().toSeconds();
	float nanosec = gplayer.queryDuration().getNanoSeconds(); 
    return sec + nanoSecToSecFrac(nanosec);
  }

  /**
   * Set how often new frames are to be read from the movie.
   * Does not actually set the speed of the movie playback,
   * that's handled by the speed() method.
   */
  public void frameRate(int ifps) {
  /*
    if (ifps <= 0) {
      System.err.println("Movie: ignoring bad frame rate of " +
                         ifps + " fps.");
    } else {
      fps = ifps;
    }
	*/
  }

  /**
   * Set a multiplier for how fast/slow the movie should be run.
   * The default is 1.0.
   * <UL>
   * <LI>speed(2) will play the movie at double speed (2x).
   * <LI>speed(0.5) will play at half speed.
   * <LI>speed(-1) will play backwards at regular speed.
   * </UL>
   */
  public void speed(float rate) {
  /*
    //rate = irate;
    try {
      movie.setRate(rate);

    } catch (StdQTException e) {
      errorMessage("speed", e);
    }
*/	
  }
  
  /**
   * Return the current time in seconds.
   * The number is a float so fractions of seconds can be used.
   */  
  public float time() {
    float sec = gplayer.queryPosition().toSeconds();
	float nanosec = gplayer.queryPosition().getNanoSeconds();
	return sec + nanoSecToSecFrac(nanosec);
  }   
  
  /**
   * Jump to a specific location (in seconds).
   * The number is a float so fractions of seconds can be used.
   */
  public void jump(float where) {
      gplayer.seek(ClockTime.fromNanos(secToNanoLong(where)));
  }  
  
  public boolean available() {
    return available;
  }
  
/**
   * Begin playing the movie, with no repeat.
   */
  public void play() {
    play = true;
	gplayer.play();
  }	
  
  /**
   * Begin playing the movie, with repeat.
   */
  public void loop() {
    play = true;
    repeat = true;
	gplayer.play();
  }

  /**
   * Shut off the repeating loop.
   */
  public void noLoop() {
    repeat = false;
  }

  /**
   * Pause the movie at its current time.
   */
  public void pause() {
    play = false;
	gplayer.pause();
  }
    
  /**
   * Stop the movie, and rewind.
   */
  public void stop() {
    play = false;	
	goToBeginning();
	gplayer.stop();
  }  

  public void read() {
	if (firstFrame) {
      super.init(movieWidth, movieHeight, RGB);
	  loadPixels();
      firstFrame = false;
    }		
	moviePixels.get(pixels);
	updatePixels();
  }
  	
  void invokeEvent(int w, int h, IntBuffer buffer) {
  
    available = true;
	movieWidth = w;
	movieHeight = h;
	moviePixels = buffer;
	
    if (play) {
    
      // Creates a movieEvent.
      if (movieEventMethod != null) {
        try {
          movieEventMethod.invoke(parent, new Object[] { this });
        } catch (Exception e) {
          System.err.println("error, disabling movieEvent() for " +
                             filename);
          e.printStackTrace();
          movieEventMethod = null;
        }
      }

      if (isDone() && repeat) {
        goToBeginning();
	  }
	  
	}
	
  }
  
  void goToBeginning() {
  
    gplayer.seek(ClockTime.fromNanos(0));
  }
  
  boolean isDone() {
	float diff = duration() - time();

    return diff < 0.033;
  }
  
  float nanoSecToSecFrac(float nanosec) {
      for (int i = 0; i < 3; i++) nanosec /= 1E3;
	  return nanosec;
  }
  
  long secToNanoLong(float sec) {
    Float f = new Float(sec * 1E9);
    return f.longValue();
  }
  
  static long INSTANCES_COUNT = 0;
 }
